/*
 *
 * Component:pbdcm
 *
 * Description:dynamic private tag handling
 * Definitions of the private elements are in defpbelems.h
 *
 *
 * Last Updated     : $Author$
 * Update Date Time : $Date$
 * RCS File Name    : $RCSfile$
 * RCS Revision     : $Revision$
 * Locked by        : $Locker$
 * Symbolic name    : $Name$
 * State            : $State$
 */
#ifndef __PBITEM_H_INCLUDED__
#define __PBITEM_H_INCLUDED__


#include <dcmtk/dcmdata/dcitem.h>
#include <dcmtk/ofstd/ofcond.h>
#include <dcmtk/dcmdata/dctk.h>
#include <dcmtk/config/osconfig.h>
#include <dcmtk/dcmnet/cond.h>
#include <map>
#ifndef WIN32
#include <syslog.h>
#endif

#define isDEBUGPBITEM true
#ifdef WIN32
#define DEBUGPBITEM printf
#else
#define DEBUGPBITEM(...) if (isDEBUGPBITEM) {openlog("PBItem", LOG_PID, LOG_LOCAL3); syslog(LOG_DEBUG, __VA_ARGS__); closelog(); }
#endif

/*
 *  DCMTK module numbers for modules which create their own error codes.
 *  Module numbers > 1023 are reserved for user code.
 */
const unsigned short OFM_pbdcm = 1051;
extern const OFCondition EC_PBPrivateTagProblem;   /* Private Tag Relocation Problem */

typedef std::map<DcmTagKey, Uint16> keymap;


/**
 * Class to handle EP private tags in items
 */
class PBItem {
	private:
		DcmItem & m_item;
		Uint16 m_privateBlock;
		bool m_isPrivateBlockReserved;
		bool m_isPrivateBlockDefined;

		bool relocateLegacyPrivateTags( bool handleCanonic = false );
		bool checkAndReservePrivateBlock();
		bool isOldBlockOurs( int oldblock );

		static void createMap();

		static bool m_isMapCreated;

	public:

		static keymap m_privateKeyMap;

		/**
		 * Makes a canonic tag from the new-element (1 byte)
		 * canonic tag looks like (f215,10xx)
		 * canonic tag is used only for internal communication!
		 * @param element  the 1 byte private element
		 * @return the canonic tag
		 */
		static DcmTag canonicTag( Uint16 element );

		/**
		 * Makes a canonic tag from a tag, if it's our private tag
		 * Also makes a new-type canonic tag from an old-type our private tag
		 * @param key  the private tag
		 * @return canonic form
		 */
		static DcmTag getCanonicKey( const DcmTag & key );

		/**
		 * @param tag  any tag
		 * @return  if it's our private creator tag
		 */
		static bool isOurPrivateCreator( const DcmTag & tag );

		/**
		 * @param tag  any tagkey
		 * @return  if it's our private tag
		 */
		static bool isOurPrivateTagKey( const DcmTagKey & tagKey );

		static DcmElement* createNewDicomElement( const DcmTagKey & tagKey );


		/**
		 * PBItem is a wrapper item around a DcmItem, which handles private block and private tags
		 */
		PBItem(DcmItem & item);

		/**
		 * Calculates the private block for this item, resolves legacy-like private tags to the new form
		 * (ba & da tags)
		 * the constructor doesn't call it, but all methods _except_ isObjectOurPrivateElement() calls it first
		 * @param handleCanonic  relocate canonic form tags, too?
		 * @return true if succeeded
		 */
		bool calcPrivateBlock( bool handleCanonic = false );

		/**
		 * Wrapper around DcmItem::findAndGetString, works for private elements
		 * @param element  the 1-byte private element
		 * @param value  the found string or remains the same (usually NULL) if not found
		 * @param searchIntoSub  search into subdirectories or not
		 * @return the condition, PBITEM_PRIVATETAGPROBLEM if the private tags couldn't be resolved
		 */
		OFCondition findAndGetPBString( Uint16 element, const char *&value, const OFBool searchIntoSub = OFFalse);

		/**
		 * Wrapper around DcmItem::findAndGetElement, works for private elements
		 * @param element  the 1-byte private element
		 * @param elementvalue  the found element or remains the same (usually NULL) if not found
		 * @param searchIntoSub  search into subdirectories or not
		 * @return the condition, PBITEM_PRIVATETAGPROBLEM if the private tags couldn't be resolved
		 */
		OFCondition findAndGetPBElement( Uint16 element, DcmElement *&elementvalue, const OFBool searchIntoSub = OFFalse);

		/**
		 * Wrapper around DcmItem::findAndGetSequence, works for private elements
		 * @param element  the 1-byte private element
		 * @param sequence  the found sequence or remains the same (usually NULL) if not found
		 * @param searchIntoSub  search into subdirectories or not
		 * @return the condition, PBITEM_PRIVATETAGPROBLEM if the private tags couldn't be resolved
		 */
		OFCondition findAndGetPBSequence( Uint16 element, DcmSequenceOfItems *&sequence, const OFBool searchIntoSub = OFFalse);

		/**
		 * Wrapper around DcmItem::findAndGetSequenceItem, works for private elements
		 * @param element  the 1-byte private element
		 * @param item  the found item or remains the same (usually NULL) if not found
		 * @param itemNum  as in DcmItem
		 * @return the condition, PBITEM_PRIVATETAGPROBLEM if the private tags couldn't be resolved
		 */
		OFCondition findAndGetPBSequenceItem( Uint16 element, DcmItem *&item, const signed long itemNum = 0);

		/**
		 * Wrapper around DcmItem::findOrCreateSequenceItem, works for private elements
		 * @param element  the 1-byte private element
		 * @param item  the found or created item
		 * @param itemNum  as in DcmItem
		 * @return the condition, PBITEM_PRIVATETAGPROBLEM if the private tags couldn't be resolved
		 */
		OFCondition findOrCreatePBSequenceItem( Uint16 element, DcmItem *&item, const signed long itemNum = 0);

		/**
		 * Wrapper around DcmItem::putAndInsertString, works for private elements
		 * @param element  the 1-byte private element
		 * @param value  the string to put
		 * @param replaceOld replace if found or not
		 * @return the condition, PBITEM_PRIVATETAGPROBLEM if the private tags couldn't be resolved
		 */
		OFCondition putAndInsertPBString( Uint16 element, const char * value, const OFBool replaceOld = OFTrue);

		/**
		 * Wrapper around DcmItem::putAndInsertUint8Array, works for private elements
		 * @param element  the 1-byte private element
		 * @param value  the array to put
		 * @param replaceOld replace if found or not
		 * @return the condition, PBITEM_PRIVATETAGPROBLEM if the private tags couldn't be resolved
		 */
		OFCondition putAndInsertPBUint8Array( Uint16 element, const Uint8 *value, const unsigned long count, const OFBool replaceOld = OFTrue);

		/**
		 * Wrapper around DcmItem::putAndInsertUint16Array, works for private elements
		 * @param element  the 1-byte private element
		 * @param value  the array to put
		 * @param replaceOld replace if found or not
		 * @return the condition, PBITEM_PRIVATETAGPROBLEM if the private tags couldn't be resolved
		 */
		OFCondition putAndInsertPBUint16Array( Uint16 element, const Uint16 *value, const unsigned long count, const OFBool replaceOld = OFTrue);

		/**
		 * Wrapper around DcmItem::insert, works for private elements
		 * it will change the tag of the element, if needed
		 * Also relocates the element, if it's in the old (ba or da) form
		 * @param elem  the DcmElement to insert
		 * @param replaceOld replace if found or not
		 * @param checkInsertOrder  as in DcmItem
		 * @return the condition, PBITEM_PRIVATETAGPROBLEM if the private tags couldn't be resolved
		 */
		OFCondition insertPB( DcmElement *elem, OFBool replaceOld = OFFalse, OFBool checkInsertOrder = OFFalse);

		/**
		 * Wrapper around DcmItem::insertEmptyElement, works for private elements
		 * @param element  the 1-byte private element
		 * @param replaceOld replace if found or not
		 * @return the condition, PBITEM_PRIVATETAGPROBLEM if the private tags couldn't be resolved
		 */
		OFCondition insertPBEmptyElement( Uint16 element, const OFBool replaceOld = OFTrue);

		/**
		 * Gives the correct private tag for the 1-byte private element
		 * @param element  the 1-byte private element
		 * @param tag  gives back the newly created tag
		 * @return the condition of the procedure
		 */
		OFCondition getTag( Uint16 element, DcmTag& tag );

		/**
		 * Gives the correct private tag for the 1-byte private element
		 * simpler to use, but no condition
		 * it will give back canonic form in case of problems
		 * @param element  the 1-byte private element
		 * @return  gives back the newly created tag
		 */
		DcmTag getTag( Uint16 element );

		/**
		 * Makes a copy of the element, relocates it if it's our private element
		 * and inserts it into the item
		 * Also relocates sub-items is item is a sequence
		 * @param oldelement  the element to copy
		 * @param replaceOld  replace if found or not
		 * @return the condition, PBITEM_PRIVATETAGPROBLEM if the private tags couldn't be resolved
		 */
		OFCondition copyPrivateElement( DcmElement& oldelement, OFBool replaceOld = OFFalse);

		/**
		 * Wrapper around DcmItem::remove, works for private elements
		 * @param ptag the 1-byte private element
		 * @return the element, which was removed, or NULL
		 */
		DcmElement * removePB( Uint16 ptag );

		/**
		 * Removes all EP private tags from the item.
		 */
		void nukePrivateTags();
			
		/**
		 * !needs calcPrivateBlock() first!
		 * before you call this for objects of an item (usually in a cycle)
		 * please call PBItem.calcPrivateBlock() to relocate private tags
		 * Gives back if this object is our given private element
		 * @param obj  the object
		 * @param element  the 1-byte private element
		 * @return true if the object is the given private element
		 */
		bool isObjectOurPrivateElement( DcmObject* obj, Uint16 element );

		/**
		 * Converts a tag to new-style format if it was our old-style private tag
		 */
		OFCondition convertTag( DcmTag& tag );
};

/**
 * Recursively relocates all private tags of an item (dataset) to the new form
 * Relocating also means handling the Private Creator tags
 * @param item  the item to relocate
 */
void handlePrivateTags( DcmItem* item );

/**
 * Recursively relocates all private tags of an item (dataset) to the new form
 * Also relocates the canonic-type new private tags
 * Relocating also means handling the Private Creator tags
 * @param item  the item to relocate
 */
void handleCanonicAndOldPrivateTags( DcmItem* item );


/**
 * Recursively relocates all private tags from a file to the new form, and saves it into another file
 * Relocating also means handling the Private Creator tags
 * @param infilename  the name of the input file
 * @param outfilename  the name of the output file
 * @return  true if succeeded
 */
bool convertFilePrivateTags( const char* infilename, const char* outfilename );

/**
 * Checks if the given group + element is our private tag (new canonic form!), and
 * changes the element part according to hte item, if neccessary
 * @param item  the item in which we work
 * @param gtag  the group
 * @param etag  the element, it can be changed!
 */
void checkAndUpdateCanonicPrivateTag( DcmItem * item, Uint16 gtag, Uint16 & etag );

/**
 * The same, but works for canonic tags instead of group + element
 * @param item  the item in which we work
 * @param tag  the tag, it can be changed!
 */
void checkAndUpdateCanonicPrivateTag( DcmItem * item, DcmTag & tag );


#endif
